#include "testlib.h"
#include <bits/stdc++.h>

using namespace std;

void choose_more_edges(int n, int m, set<pair<int, int>>& edges, bool only_right = false) {
    vector<pair<int, int>> available;
    for (int u = 0; u < n; ++u) {
        for (int v = u + 1; v < n; ++v) {
            if (!edges.count({u, v}) && !edges.count({v, u})) {
                available.emplace_back(u, v);
            }
        }
    }
    ensuref(available.size() >= m - (int)edges.size(), "too many edges");
    vector<int> idx = rnd.distinct(m - (int)edges.size(), 0, (int)available.size() - 1);
    for (int i : idx) {
        auto [u, v] = available[i];
        if (only_right || rnd.next(0, 1)) {
            edges.emplace(u, v);
        } else {
            edges.emplace(v, u);
        }
    }
}

void output(int n, int m, set<pair<int, int>>& edges) {
    ensuref(m == edges.size(), "wanted %d, generated %d edges", m, (int)edges.size());
    printf("%d %d\n", n, m);
    for (auto [u, v] : edges) {
        printf("%d %d\n", u + 1, v + 1);
    }
}

int main(int argc, char *argv[]) {
    registerGen(argc, argv, 1);

    suppressEnsureNoUnusedOpts();

    bool exact = opt<bool>("exact", false);
    int N = opt<int>("n", 2000);
    int M = opt<int>("m", 2000);
    bool dag = opt<bool>("dag", false);
    bool scc = opt<bool>("scc", false);
    bool hamiltonian = opt<bool>("hamiltonian", false);

    ensuref(!dag || !scc, "-dag and -scc cannot be used together");
    ensuref(!hamiltonian || dag || scc, "-hamiltonian cannot be used without -dag or -scc");

    int n, m;
    set<pair<int, int>> edges;

    if (exact) {
        n = N, m = M;
    }

    if (!dag && !scc) {
        if (!exact) {
            n = rnd.next(2, N);
            m = rnd.next(1, min(M, n * (n - 1) / 2));
        }

        choose_more_edges(n, m, edges);
    } else if (dag) {
        if (!exact) {
            n = rnd.next(2, N);
            if (!hamiltonian) {
                m = rnd.next(1, min(M, n * (n - 1) / 2));
            } else {
                m = rnd.next(n - 1, min(M, n * (n - 1) / 2));
            }
        }

        if (hamiltonian) {
            for (int i = 0; i < n - 1; ++i) {
                edges.emplace(i, i + 1);
            }
        }

        choose_more_edges(n, m, edges, true);
    } else if (scc) {
        if (!exact) {
            n = rnd.next(3, N);
            m = rnd.next(n, min(M, n * (n - 1) / 2));
        }

        if (hamiltonian) {
            for (int i = 0; i < n - 1; ++i) {
                edges.emplace(i, i + 1);
            }
            edges.emplace(n - 1, 0);
        } else {
            int blocks;
            if (has_opt("scc-blocks")) {
                blocks = min(opt<int>("scc-blocks"), min(m - n + 1, n - 2));
            } else {
                // what the hell
                blocks = rnd.wnext(1, min(m - n + 1, n - 2), -max(1, int(sqrt(n)) / 2));
            }
            vector<int> idx = rnd.distinct(blocks - 1, 3, n - 1);
            sort(idx.begin(), idx.end());
            idx.push_back(n);
            vector<int> perm = rnd.perm(n);
            for (int i = 0; i < idx[0] - 1; ++i) {
                edges.emplace(perm[i], perm[i + 1]);
            }
            edges.emplace(perm[idx[0] - 1], perm[0]);
            for (int b = 0; b < blocks - 1; ++b) {
                int from, to;
                if (idx[b + 1] - idx[b] > 1) {
                    from = rnd.next(0, idx[b] - 1);
                    to = rnd.next(0, idx[b] - 1);
                } else {
                    auto ends = rnd.distinct(2, 0, idx[b] - 1);
                    from = ends[0];
                    to = ends[1];
                }
                edges.emplace(perm[from], perm[idx[b]]);
                for (int i = idx[b]; i < idx[b + 1] - 1; ++i) {
                    edges.emplace(perm[i], perm[i + 1]);
                }
                edges.emplace(perm[idx[b + 1] - 1], perm[to]);
            }
        }

        choose_more_edges(n, m, edges);
    }

    output(n, m, edges);
}
